@@ -33,6 +33,7 @@ typedef enum { BINARY, CONFIG } Format;
 const char *ws = " \t";
 char *device;
 uint16_t delay = 0;
+bool preserve_state = false;
 
 //------------------------------------------------------------------------------
 // Utility functions for parsing
@@ -41,22 +42,24 @@ uint16_t delay = 0;
 static uint8_t parseAction(char* str) {
   if(str == NULL) return ACTION_NONE;
   
-  if(strncasecmp(str, "set",   3) == 0) return ACTION_SET;
-  if(strncasecmp(str, "clear", 5) == 0) return ACTION_CLEAR;
-  if(strncasecmp(str, "inv",   3) == 0) return ACTION_INVERT;
-  if(strncasecmp(str, "inc",   3) == 0) return ACTION_INCREASE;
-  if(strncasecmp(str, "dec",   3) == 0) return ACTION_DECREASE;
-  if(strncasecmp(str, "tri",   3) == 0) return ACTION_TRISTATE;
-  if(strncasecmp(str, "sleep", 5) == 0) return ACTION_SLEEP_SHORT;
-  if(strncasecmp(str, "exec",  4) == 0) return ACTION_EXEC;
-  if(strncasecmp(str, "meta",  4) == 0) return ACTION_DEFINE_META;
-  if(strncasecmp(str, "down",  2) == 0) return ACTION_KEY_DOWN;
-  if(strncasecmp(str, "up",    2) == 0) return ACTION_KEY_UP;
-  if(strncasecmp(str, "type",  4) == 0) return ACTION_TYPE;  
-  if(strncasecmp(str, "boot",  4) == 0) return ACTION_BOOT;    
-  if(strncasecmp(str, "swap",  4) == 0) return ACTION_SWAP;  
-  if(strncasecmp(str, "press", 5) == 0) return ACTION_KEY_PRESS;
-  if(strncasecmp(str, "using", 5) == 0) return ACTION_DEFINE_SWITCH;    
+  if(strncasecmp(str, "set",      3) == 0) return ACTION_SET;
+  if(strncasecmp(str, "clear",    5) == 0) return ACTION_CLEAR;
+  if(strncasecmp(str, "inv",      3) == 0) return ACTION_INVERT;
+  if(strncasecmp(str, "inc",      3) == 0) return ACTION_INCREASE;
+  if(strncasecmp(str, "dec",      3) == 0) return ACTION_DECREASE;
+  if(strncasecmp(str, "tri",      3) == 0) return ACTION_TRISTATE;
+  if(strncasecmp(str, "sleep",    5) == 0) return ACTION_SLEEP_SHORT;
+  if(strncasecmp(str, "exec",     4) == 0) return ACTION_EXEC;
+  if(strncasecmp(str, "meta",     4) == 0) return ACTION_DEFINE_META;
+  if(strncasecmp(str, "down",     2) == 0) return ACTION_KEY_DOWN;
+  if(strncasecmp(str, "up",       2) == 0) return ACTION_KEY_UP;
+  if(strncasecmp(str, "type",     4) == 0) return ACTION_TYPE;  
+  if(strncasecmp(str, "boot",     4) == 0) return ACTION_BOOT;    
+  if(strncasecmp(str, "swap",     4) == 0) return ACTION_SWAP;  
+  if(strncasecmp(str, "press",    5) == 0) return ACTION_KEY_PRESS;
+  if(strncasecmp(str, "using",    5) == 0) return ACTION_DEFINE_SWITCH;
+  if(strncasecmp(str, "save",     4) == 0) return ACTION_SAVE_STATE;
+  if(strncasecmp(str, "restore",  7) == 0) return ACTION_RESTORE_STATE;        
   return ACTION_NONE;
 }
 
@@ -152,6 +155,21 @@ static bool parseData(char *str, uint8_t *data) {
   return result;    
 }
 
+static bool parsePolicy(Command* command, char *str) {
+  bool result = true;
+
+  if(strcmp(str, "0") == 0) {
+    command->policy = POLICY_EVEN;
+  }
+  else if(strcmp(str, "1") == 0) {
+    command->policy = POLICY_ODD;
+  }
+  else {
+    result = false;
+  }
+  return result;
+}
+
 //------------------------------------------------------------------------------
 
 static uint32_t parseDuration(char *str) {
@@ -262,7 +280,7 @@ bool Config_parse(Config* self, FILE* in) {
     if(line[strlen(line)-1] == '\r') {
       line[strlen(line)-1] = '\0';
     }
-    
+
     // check if this command shall be bound to a key
     if((colon = strstr(line, ":")) != NULL) {
       keyspec = line;
@@ -359,10 +377,16 @@ bool Command_parse(Command* self, char* spec) {
   uint8_t data;
   uint16_t index;
   uint32_t value;
+  bool has_policy = false;
   int i = 0;
   
   StringList_append_tokenized(words, spec, ws);
 
+  if(parsePolicy(self, StringList_get(words, i))) {
+    i++;
+    has_policy = true;
+  }
+  
   if((self->action = parseAction(StringList_get(words, i++))) == ACTION_NONE) {
     fprintf(stderr, "error: '%s': invalid command\n", StringList_get(words, i-1));
     goto error;
@@ -371,6 +395,10 @@ bool Command_parse(Command* self, char* spec) {
   if(self->action == ACTION_TYPE) {
     str = spec + strcspn(spec, ws) + 1;
 
+    if(has_policy) {
+      str = str + strcspn(str, ws) + 1;
+    }
+
     if(!Config_has_string(config, str, &index)) {
       index = Config_add_string(config, str);
     }
@@ -509,16 +537,18 @@ bool Command_parse(Command* self, char* spec) {
 void Config_write(Config *self, FILE *out) {
   fputc(CONFIG_MAGIC[0], out);
   fputc(CONFIG_MAGIC[1], out);
+
+  State_write(self->state, out);
   
-  for(int i=0; i<self->size; i++) {
+  for(int i=0; i<self->num_bindings; i++) {
     Binding_write(self->bindings[i], out);
   }
-  for(int i=0; i<self->_size; i++) {
+  for(int i=0; i<self->num_strings; i++) {
     fputc(KEY_STRING, out);
     fputs(self->strings[i], out);
     fputc('\0', out);
   }
-  for(int i=0; i<self->__size; i++) {
+  for(int i=0; i<self->num_longs; i++) {
     fputc(KEY_LONG, out);
     fputc((self->longs[i] >> 0) & 0xff, out);
     fputc((self->longs[i] >> 8) & 0xff, out);
@@ -532,9 +562,9 @@ void Config_write(Config *self, FILE *out) {
 
 void Binding_write(Binding *self, FILE* out) {
   Key_write(self->key, out);
-  fputc(self->size, out);
+  fputc(self->num_commands, out);
   
-  for(int i=0; i<self->size; i++) {
+  for(int i=0; i<self->num_commands; i++) {
     Command_write(self->commands[i], out);
   }
 }
@@ -550,17 +580,38 @@ void Key_write(uint8_t key, FILE* out) {
 void Command_write(Command *self, FILE* out) {
   uint8_t action = self->action;
   action |= self->port<<7;
+  action |= self->policy;
   fputc(action, out);
   fputc(self->mask, out);
   fputc(self->data, out);
 }
 
+//------------------------------------------------------------------------------
+
+void State_write(State *self, FILE* out) {
+  fputc(self->ddra, out);
+  fputc(self->porta, out);
+  fputc(self->ddrb, out);
+  fputc(self->portb, out);
+}
+
+//------------------------------------------------------------------------------
+
+bool State_fetch(State *self) {
+  int result;
+  
+  if((result = usb_receive(device, KEYMAN64_STATE, 0, (uint8_t*) self, sizeof(State))) < 0) {
+    fprintf(stderr, "error: could send usb control message: %s\n", libusb_strerror(result));
+  }
+  return result;
+}
+
 //------------------------------------------------------------------------------
 // Functions for writing canonical config file format
 //------------------------------------------------------------------------------
 
 void Config_print(Config *self, FILE* out) {
-  for(int i=0; i<self->size; i++) {
+  for(int i=0; i<self->num_bindings; i++) {
     Binding_print(self->bindings[i], out);
   }
 }
@@ -569,7 +620,7 @@ void Config_print(Config *self, FILE* out) {
 
 void Binding_print(Binding *self, FILE* out) {
   
-  for(int i=0; i<self->size; i++) {
+  for(int i=0; i<self->num_commands; i++) {
     if(self->key != KEY_IMMEDIATE) {
       Key_print(self->key, out);
     }
@@ -606,9 +657,19 @@ void Command_print(Command *self, FILE* out) {
   case ACTION_BOOT:          action = "boot";     break;                
   case ACTION_SWAP:          action = "swap";     break;
   case ACTION_KEY_PRESS:     action = "press";    break;
-  case ACTION_DEFINE_SWITCH: action = "using";    break;                        
+  case ACTION_DEFINE_SWITCH: action = "using";    break;
+  case ACTION_SAVE_STATE:    action = "save";     break;
+  case ACTION_RESTORE_STATE: action = "restore";  break;                                
   };
 
+  if(self->policy == POLICY_EVEN) {
+    fprintf(out, "0 ");
+  }
+  
+  if(self->policy == POLICY_ODD) {
+    fprintf(out, "1 ");
+  }
+  
   if(self->action == ACTION_TYPE) {
     uint16_t index = self->mask | (self->data << 8);
     fprintf(out, "%s ", action);
@@ -634,7 +695,9 @@ void Command_print(Command *self, FILE* out) {
     }
   }
 
-  if(self->action == ACTION_BOOT) {
+  if(self->action == ACTION_BOOT ||
+     self->action == ACTION_SAVE_STATE ||
+     self->action == ACTION_RESTORE_STATE) {
     fprintf(out, "%s", action);
   }
 
@@ -732,18 +795,19 @@ int main(int argc, char **argv) {
   strcpy(device, default_device);
   
   struct option options[] = {
-    { "help",    no_argument,       0, 'h' },
-    { "version", no_argument,       0, 'v' },
-    { "device",  required_argument, 0, 'd' },
-    { "keys",    no_argument,       0, 'k' },
-    { "delay",   required_argument, 0, 'D' },
+    { "help",     no_argument,       0, 'h' },
+    { "version",  no_argument,       0, 'v' },
+    { "device",   required_argument, 0, 'd' },
+    { "keys",     no_argument,       0, 'k' },
+    { "delay",    required_argument, 0, 'D' },
+    { "preserve", no_argument,       0, 'p' },
     { 0, 0, 0, 0 },
   };
   int option, option_index;
 
 
   while(1) {
-    option = getopt_long(argc, argv, "hvd:kD:", options, &option_index);
+    option = getopt_long(argc, argv, "hvd:kD:p", options, &option_index);
 
     if(option == -1)
       break;
@@ -774,6 +838,10 @@ int main(int argc, char **argv) {
       delay = strtol(optarg, NULL, 0);
       break;
 
+    case 'p':
+      preserve_state = true;
+      break;      
+      
     case '?':
     case ':':
       goto done;
@@ -826,14 +894,6 @@ int command(int argc, char **argv) {
   uint8_t *data = NULL;
   uint16_t size = 0;
   
-  libusb_device_handle *handle = NULL;
-  DeviceInfo info;
-  
-  if((result = libusb_init(NULL)) < 0) {
-    fprintf(stderr, "error: could not initialize libusb-1.0: %s\n", libusb_strerror(result));
-    goto done;
-  }
-
   if(argc) {
 
     if((in = fopen(argv[0], "rb")) == NULL) {
@@ -872,29 +932,13 @@ int command(int argc, char **argv) {
   fmemupdate(out, data, size);  
   fclose(out);
 
-  info.vid = KEYMAN64_VID;
-  info.pid = KEYMAN64_PID;
-
-  usb_lookup(device, &info);
-  handle = usb_open(NULL, &info);
-  
-  if(handle == NULL) {
-    fprintf(stderr, "error: could not open %s\n", device);
-    goto done;
-  }
-
-  if((result = usb_send(handle, KEYMAN64_CTRL, delay, data, size)) < 0) {
+  if((result = usb_send(device, KEYMAN64_CTRL, delay, data, size)) < 0) {
     fprintf(stderr, "error: could send usb control message: %s\n", libusb_strerror(result));
     goto done;
   }
   result = EXIT_SUCCESS;
   
  done:
-  if(handle != NULL) {
-    libusb_close(handle);
-  }
-  libusb_exit(NULL);
-  
   if(config != NULL) {
     Config_free(config);
   }
@@ -947,6 +991,12 @@ int convert(int argc, char **argv) {
   
   if(Config_read(config, in) || Config_parse(config, in)) {
 
+    if(preserve_state) {
+      if(!State_fetch(config->state)) {
+        fprintf(stderr, "warning: could not fetch saved state from device\n");
+      }        
+    }
+
     output_format == BINARY ?
       Config_write(config, out) :
       Config_print(config, out);
@@ -979,20 +1029,21 @@ void usage(void) {
   printf("\n");
   printf("Usage:\n");
   printf("      keyman64 <options>\n");
-  printf("      keyman64 convert [<infile>|-] [<outfile>|-]\n");
+  printf("      keyman64 <options> convert [<infile>|-] [<outfile>|-]\n");
   printf("      keyman64 [<options>] <command>\n");
   printf("      keyman64 [<options>] [<file>]\n");    
   printf("\n");
   printf("  Options:\n");
-  printf("           -v, --version : print version information\n");
-  printf("           -h, --help    : print this help text\n");
+  printf("           -v, --version  : print version information\n");
+  printf("           -h, --help     : print this help text\n");
 #if linux
-  printf("           -d, --device  : specify usb device (default: /dev/keyman64)\n");
+  printf("           -d, --device   : specify usb device (default: /dev/keyman64)\n");
 #elif windows
-  printf("           -d, --device  : specify usb device (default: usb)\n");
+  printf("           -d, --device   : specify usb device (default: usb)\n");
 #endif
-  printf("           -D, --delay   : delay in ms between commands\n");  
-  printf("           -k, --keys    : list key names and synonyms\n");
+  printf("           -D, --delay    : delay in ms between commands\n");  
+  printf("           -k, --keys     : list key names and synonyms\n");
+  printf("           -p, --preserve : preserve saved state during convert\n");  
   printf("\n");
   printf("  Conversion arguments:\n");
   printf("           <infile>  : input file, format is autodetected\n");
